# Copyright 2022 The MOGPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from .gps import GPLVM

__version__ = "0.0.2"
__authors__ = "Daniel Dodd, Thomas Pinder"
__emails__ = "d.dodd1@lancaster.ac.uk, tompinder@live.co.uk"
__license__ = "Apache 2.0"
__uri__ = "https://github.com/Daniel-Dodd/jax_linear_operator"
__description__ = "A JaxLinOp library."
__contributors__ = "https://github.com/Daniel-Dodd/jax_linear_operator/graphs/contributors"

__all__ = [
    "GPLVM", 
]


# Copyright 2022 The MOGPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import Dict, Callable
from jaxtyping import Float, Array

import jax.numpy as jnp
from jax import vmap
from jax.random import KeyArray
import distrax as dx

import gpjax as gpx
from gpjax.utils import concat_dictionaries
from jaxutils import PyTree, Dataset

class AbstractGP(PyTree):
    pass

class Prior(AbstractGP):
    pass

class AbstractPosterior(AbstractGP):
    pass

class ConjugatePosterior(AbstractPosterior):
    pass

class NonConjugatePosterior(AbstractPosterior):
    pass


class GPLVM(PyTree):
    """A Gaussian Process Latent Variable Model (GPLVM)."""

    def __init__(
        self,
        latent_process: gpx.gps.ConjugatePosterior,
        latent_dim: int,
        ) -> None:

        self.latent_process = latent_process
        self.latent_dim = latent_dim
    
    

    def _initialise_params(self, key: KeyArray) -> Dict:
        """Initialise the GP's parameter set.

        Args:
            key (KeyArray): The random key to use for initialisation.

        Returns:
            Dict: The initial parameter set.
        """

        latent_process = self.latent_process
        n_data = latent_process.likelihood.num_datapoints

        latent_prior = dx.MultivariateNormalDiag(
            loc=jnp.zeros(self.latent_dim), 
            scale_diag=jnp.ones(self.latent_dim)
        )

        model_params = self.latent_process._initialise_params(key)

        X = latent_prior.sample(seed=key, sample_shape=(n_data,))

        return concat_dictionaries(model_params, {"latent": X})

    def marginal_log_likelihood(
        self,
        train_data: Dataset,
        priors: Dict = None,
        negative: bool = False,
    ) -> Callable[[Dict], Float[Array, "1"]]:
        """Compute the marginal log likelihood of the model.
        Args:
            train_data (Dataset): The training dataset.
            priors (dict): The priors to apply to the training dataset.
            negative (bool): Whether to return the negative log likelihood.
        Returns:
            Callable[[Dict], Float[Array, "1"]]: A function that takes a dictionary of parameters and returns the marginal log likelihood.
        """

        if not train_data.is_unsupervised():
            raise ValueError("Data must be unsupervised.")

        def _single_mll(y: Float[Array, "N Q"] , params: Dict):
            X = params["latent"]
            D = Dataset(X, y)
            
            mll = self.latent_process.marginal_log_likelihood(
                D, priors
            )
            return mll(params)

        constant = jnp.array(-1.0) if negative else jnp.array(1.0)
        y = train_data.y

        def mll(params):
            return constant * jnp.sum(
                vmap(_single_mll, in_axes=(0, None))(
                    jnp.transpose(y), params
                )
            )

        return mll

__all__ = [
    "GPLVM",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
import os
import re
from setuptools import find_packages, setup

NAME = "mogpjax"

CLASSIFIERS = [
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: Apache License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3",
]

INSTALL_REQUIRES = [
    "jax",
    "jaxlib",
    "jaxtyping",
    "distrax",
    "jaxutils",
    "jaxlinop",
    "gpjax>=0.5.4"
]

EXTRA_REQUIRE = {
    "dev": [
        "pytest",
        "pre-commit",
        "pytest-cov",
    ],
}

GLOBAL_PATH = os.path.dirname(os.path.realpath(__file__))


def read(*local_path: str) -> str:
    """Read a file, given a local path.

    Args:
        *local_path (str): The local path to the file.

    Returns:
        str: The contents of the file.
    """
    with codecs.open(os.path.join(GLOBAL_PATH, *local_path), "rb", "utf-8") as f:
        return f.read()


# Read `__init__` file:
init_local_path = os.path.join(NAME, "__init__.py")
init_file = read(init_local_path)


def find_meta(meta: str) -> str:
    """Extract `__*meta*__` from the `__init__.py` file. This is useful for extracting __version__, __author__, etc.

    Args:
        meta (str): The meta to extract.

    Returns:
        str: The meta.
    """

    matches = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta), init_file, re.M
    )
    if matches:
        return matches.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


if __name__ == "__main__":
    setup(
        name=NAME,
        version=find_meta("version"),
        author=find_meta("authors"),
        author_email=find_meta("emails"),
        maintainer=find_meta("authors"),
        maintainer_email=find_meta("emails"),
        url=find_meta("uri"),
        license=find_meta("license"),
        description=find_meta("description"),
        long_description=read("README.md"),
        long_description_content_type="text/markdown",
        packages=find_packages(".", exclude=["tests"]),
        package_data={NAME: ["py.typed"]},
        include_package_data=True,
        python_requires=">=3.6",
        install_requires=INSTALL_REQUIRES,
        extras_require=EXTRA_REQUIRE,
        zip_safe=True,
    )
